#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "AkelDLL.h"


//Defines
#ifndef EM_SETTYPOGRAPHYOPTIONS
  #define EM_SETTYPOGRAPHYOPTIONS (WM_USER + 202)
#endif
#ifndef EM_GETTYPOGRAPHYOPTIONS
  #define EM_GETTYPOGRAPHYOPTIONS (WM_USER + 203)
#endif
#ifndef TO_ADVANCEDTYPOGRAPHY
  #define TO_ADVANCEDTYPOGRAPHY 1
#endif
#ifndef TO_SIMPLELINEBREAK
  #define TO_SIMPLELINEBREAK 2
#endif

//Functions prototypes
void Justify(HWND hWnd, WORD wAlignment);
void SaveCharScroll(HWND hWnd, RECT *rc, int *nBeforeChar);
void RestoreCharScroll(HWND hWnd, RECT *rc, int *nBeforeChar);
void SetTextSel(HWND hWnd, int nMin, int nMax);
void InitCommon(PLUGINDATA *pd);

//Global variables
HWND hMainWnd;
HWND hWndEdit;
BOOL bOldRichEdit;
BOOL bInitCommon=FALSE;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(-1, -1, -1, -1);
  pv->pPluginName="Justify";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (!bOldRichEdit)
  {
    int nTypography;

    nTypography=SendMessage(hWndEdit, EM_GETTYPOGRAPHYOPTIONS, 0, 0);
    if (!(nTypography & TO_ADVANCEDTYPOGRAPHY))
      SendMessage(hWndEdit, EM_SETTYPOGRAPHYOPTIONS, (nTypography & ~TO_SIMPLELINEBREAK) | TO_ADVANCEDTYPOGRAPHY, (nTypography & ~TO_SIMPLELINEBREAK) | TO_ADVANCEDTYPOGRAPHY);
    Justify(hWndEdit, PFA_JUSTIFY);
  }
}

void Justify(HWND hWnd, WORD wAlignment)
{
  CHARRANGE crInitial;
  PARAFORMAT2 pf;
  EDITINFO ei;
  RECT rcEdit;
  int nFirstVisibleChar;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    //Reserve state
    SendMessage(hWnd, WM_SETREDRAW, FALSE, 0);
    SaveCharScroll(hWnd, &rcEdit, &nFirstVisibleChar);

    pf.cbSize=sizeof(PARAFORMAT2);
    pf.dwMask=PFM_ALIGNMENT;
    pf.wAlignment=wAlignment;

    SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&crInitial);
    if (crInitial.cpMin == crInitial.cpMax)
      SetTextSel(hWnd, 0, -1);
    SendMessage(hWnd, EM_SETPARAFORMAT, 0, (LPARAM)&pf);
    SetTextSel(hWnd, crInitial.cpMin, crInitial.cpMin);
    if (crInitial.cpMin != crInitial.cpMax)
      SetTextSel(hWnd, crInitial.cpMin, crInitial.cpMax);

    //Restore state
    SendMessage(hMainWnd, AKD_SETMODIFY, (WPARAM)ei.hWndEdit, ei.bModified);
    RestoreCharScroll(hWnd, &rcEdit, &nFirstVisibleChar);
    SendMessage(hWnd, WM_SETREDRAW, TRUE, 0);
    InvalidateRect(hWnd, NULL, FALSE);
  }
}

void SaveCharScroll(HWND hWnd, RECT *rc, int *nBeforeChar)
{
  SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)rc);
  *nBeforeChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)rc);
}

void RestoreCharScroll(HWND hWnd, RECT *rc, int *nBeforeChar)
{
  int nAfterChar;
  int nBeforeLine;
  int nAfterLine;

  nAfterChar=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)rc);

  if (*nBeforeChar != nAfterChar)
  {
    nBeforeLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, *nBeforeChar);
    nAfterLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nAfterChar);

    if (nBeforeLine != nAfterLine)
    {
      if (bOldRichEdit)
      {
        SendMessage(hWnd, EM_LINESCROLL, 0, -1);
        SendMessage(hWnd, EM_LINESCROLL, 0, nBeforeLine);
      }
      else SendMessage(hWnd, EM_LINESCROLL, 0, nBeforeLine - nAfterLine);
    }
  }
}

void SetTextSel(HWND hWnd, int nMin, int nMax)
{
  CHARRANGE cr;

  cr.cpMin=nMin;
  cr.cpMax=nMax;
  SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  hWndEdit=pd->hWndEdit;
  bOldRichEdit=pd->bOldRichEdit;
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
  }
  return TRUE;
}
